from didipack.utils_didi.smart_algebra import smart_eigenvalue_decomposition_covariance
import numpy as np

def run_efficient_ridge(
        signals: np.ndarray,
        labels: np.ndarray,
        shrinkage_list: list,
        ):
    eigval, eigvec = smart_eigenvalue_decomposition_covariance(features=signals)
    mu = (signals * labels.reshape(-1, 1)).mean(0).reshape(-1, 1)
    multiplied = np.matmul(eigvec.T, mu)  # for a window of 2000 and 4000 signals, the cost is 4000 * 4000
    normalized = np.concatenate([(1 / (eigval + z)).reshape(-1, 1) * multiplied for z in shrinkage_list], axis=1)
    # here it is subtle as the dimension of eigvec might be lower than that of beta !!!
    # but normalized has the right dimension !!
    betas = np.matmul(eigvec, normalized).T
    return betas


def smart_eff_weights_no_q(f_ins, z_grid, n_):
    """
    Compute efficient weights from in-sample factor returns using smart_inverse_times_vector

    :param f_ins: the signals coming in this function are equivalent to (R*S).groupby(date).sum(), except for line 21 and alike where it's (R*S).groupby(date).mean()
    (as discussed I multiplied the old factors by n_(
    :param z_grid: gris of shrinkage parameters
    :param n_: a vector of dim T by 1. Containing for each date in the rolling window the number of observations n.
    if n_ is None, I am pricing the fama factors.
    :return:
    :rtype:
    """

    if n_ is not None: # in the current setup, if n_ is None it means we are with the fama french.
        f_ins = f_ins * np.sqrt(n_)

    t_, p1 = f_ins.shape
    if p1 > t_:
        b_t_tilde = f_ins @ f_ins.T / t_
        eigenvalues_b_t_tilde, eigenvectors_b_t_tilde = np.linalg.eigh(b_t_tilde)
        if (eigenvalues_b_t_tilde < 0).any():
            print("Negative eigenvalues encountered! Check f_ins where there are factor returns are identical")
            print("For now, use the slow traditional method")
            eigenvalues_b_t_, eigenvectors_b_t = np.linalg.eigh((f_ins.T @ f_ins) / t_) # removed /n_.
            eigenvectors_b_t = eigenvectors_b_t[:, -t_:]
        else:
            eigenvectors_b_t = (f_ins.T @ eigenvectors_b_t_tilde) * (eigenvalues_b_t_tilde ** (-0.5)).reshape(1,-1) / np.sqrt(t_) # TODO ask semyon to checkif I was right to remove n as well
    else:
        eigenvalues_b_t_tilde, eigenvectors_b_t_tilde = np.linalg.eigh((f_ins.T @ f_ins) / t_) # Todo I'm assuming it's ok but here too, I removed the n_
        eigenvectors_b_t = eigenvectors_b_t_tilde

    if n_ is None:
        f_mean = f_ins.mean(axis=0).reshape(-1, 1)
    else:
        f_mean = (f_ins*np.sqrt(n_)).mean(axis=0).reshape(-1, 1)
    eff_weights = np.array([smart_inverse_times_vector(z_, eigenvectors_b_t, eigenvalues_b_t_tilde, f_mean) for z_ in z_grid]).T # Todo here too I removed the divided by n_

    return eff_weights.squeeze() # , eigenvalues_b_t_tilde, eigenvectors_b_t_tilde


def smart_inverse_times_vector(z_: float,
                               eigenvectors_b_t: np.ndarray,
                               eigenvalues_without_zeros: np.ndarray,
                               vector: np.ndarray):
    """
    This is a function that uses spectral decomposition to do (zI+B)^{-1}v = U (z+D)^{-1}U' v+(I-UU') z^{-1} v
    Parameters
    ----------
    z_
    eigenvectors_b_t
    eigenvalues_without_zeros
    vector

    Returns
    -------

    """
    mult1 = eigenvectors_b_t.T @ vector

    multiplied = (1 / (z_ + eigenvalues_without_zeros)).reshape(-1, 1) * mult1
    term1 = eigenvectors_b_t @ multiplied
    # this is a correction accounting for missing zero eigenvalues
    term2 = (1 / z_) * (vector - eigenvectors_b_t @ mult1)
    return term1 + term2


from didipack.latex_table import *
from didipack.parameters import Params
from didipack.latex_ressource import *
from didipack.latex_paper import *
from didipack.plot_function import *
from .utils_didi.pandasplus import PandasPlus
from .utils_didi.plot import PlotPlus
from .utils_didi.general import General
from .create_rf.RandomFeaturesGenerator import RandomFeaturesGenerator,RandomFeaturesSpecs
from .create_rf.gen_rf_factors import generate_random_features_in_blocks,get_block_sizes




import sys
from didipack.parameters import Params
import pandas as pd
import re
from matplotlib import pyplot as plt
from typing import Tuple
from abc import ABC, abstractmethod


class BaseTrainer(ABC):
    def __init__(self, par: Params):
        self.par = par
        self.verbose = True

    def train_at_time_t(self, x: pd.DataFrame, y: pd.DataFrame,ids:pd.Series,times:pd.Series, t_index_for_split, hyper_parameters_set: dict = {}):
        x_train, y_train, x_val, y_val, x_test, y_test = self._split_data(x, y,ids=ids, times=times, t_index=t_index_for_split)

        # first we need to run the validation procedure.
        best_hyper_params = self._validation_procedure(x_train, y_train, x_val, y_val, hyper_parameters_set)

        # final training
        x = pd.concat([x_train, x_val], axis=0)
        y = pd.concat([y_train, y_val], axis=0)
        y_oos, y_ins = self._train_model_and_predict(x_train=x, y_train=y, x_test=x_test,hyper_params=best_hyper_params)
        y_test['pred'] = y_oos
        if self.par.train.save_ins:
            y['pred'] = y_ins
        else:
            y = pd.DataFrame()
        return y_test, y

    def _split_and_index(self, x: pd.DataFrame, y: pd.DataFrame, ids: pd.Series, times: pd.Series, start, end) -> Tuple[
        pd.DataFrame, pd.DataFrame]:
        indices = (times >= start) & (times < end)
        x_sub = x.loc[indices, :]
        y_sub = y.loc[indices, :]
        ids_sub = ids.loc[indices]
        times_sub = times.loc[indices]
        y_sub = y_sub.set_index([times_sub, ids_sub])
        return x_sub, y_sub

    def _split_data(self, x: pd.DataFrame, y: pd.DataFrame, ids: pd.Series, times: pd.Series, t_index) \
            -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:

        if self.par.train.T_train < 0:
            start_train = t_index - self.par.train.T_train
        else:
            start_train = t_index - self.par.train.T_train


        end_train = t_index - self.par.train.T_val
        end_test = t_index + self.par.train.testing_window

        x_train, y_train = self._split_and_index(x, y, ids, times, start_train, end_train)
        x_val, y_val = self._split_and_index(x, y, ids, times, end_train, t_index)
        x_test, y_test = self._split_and_index(x, y, ids, times, t_index, end_test)



        return x_train, y_train, x_val, y_val, x_test, y_test

    def _train_model_and_predict(self, x_train: pd.DataFrame, y_train: pd.DataFrame, x_test: pd.DataFrame,
                                 hyper_params: dict):
        self._train_model(x=x_train, y=y_train, hyper_params=hyper_params)
        y_oos = self._predict(x_test)
        if self.par.train.save_ins:
            y_ins = self._predict(x_train)
        else:
            y_ins = None
        return y_oos, y_ins

    @abstractmethod
    def _validation_procedure(self, x_train: pd.DataFrame, y_train: pd.DataFrame, x_val: pd.DataFrame,
                              y_val: pd.DataFrame, hyper_parameters_set : dict) -> dict:
        best_hypper = {}
        return best_hypper

    @abstractmethod
    def _predict(self, x):
        y = x.sum(1)
        return y

    @abstractmethod
    def _train_model(self, x: pd.DataFrame, y: pd.DataFrame, hyper_params: dict):
        if self.verbose:
            print('Model trained', flush=True)


if __name__ == "__main__":
    try:
        grid_id = int(sys.argv[1])
        model_id = int(sys.argv[2])
        print('Running with args', grid_id, flush=True)
    except:
        print('Debug mode on local machine', flush=True)
        grid_id = 0
        model_id = 5




import numpy as np
import pandas as pd
import tqdm
from abc import ABC, abstractmethod

class PandasPlus:

    @staticmethod
    def get_memory_usage_in_gb(df:pd.DataFrame):
        total_mem_usage_bytes = df.memory_usage(deep=True).sum()
        total_mem_usage_gb = total_mem_usage_bytes / 1024 ** 3
        return total_mem_usage_gb

    @staticmethod
    def winzorize_series(s, p=5):
        """
        Winzorizes a pandas Series by p percent on both sides of the distribution.

        Parameters:
        - s: The pandas Series to winzorize.
        - p: The percentage to winzorize on both sides. Defaults to 5%.

        Returns:
        - The winzorized Series.
        """
        lower = s.quantile(p / 100)
        upper = s.quantile(1 - p / 100)

        return s.clip(lower=lower, upper=upper)

    @staticmethod
    def get_ym(s:pd.Series):
        '''
        get yms from pandas
        :param s:
        :return:
        '''
        return s.dt.year*100 + s.dt.month

    import pandas as pd

    def get_yw(s: pd.Series) -> pd.Series:
        '''
        Get unique year-week identifier from a datetime pandas Series.

        :param s: A datetime-like pandas Series
        :return: A Series with unique year-week identifiers
        '''
        return s.dt.year * 100 + s.dt.week

    @staticmethod
    def weighted_sum_pandas_series(w: pd.Series, target: pd.Series, gb: pd.Series):
        return (target * w).groupby(gb).sum() / w.groupby(gb).sum()


    @staticmethod
    def normalize_by_gp_rank(s: pd.Series, gp: pd.Series):
        '''
        Rank columns in a way that sums to 1.


        :param df:
        :param col_list:
        :param groupby_name:
        :return:
        '''
        t = s.groupby(gp.values).transform('rank') - 1
        t_n = s.groupby(gp.values).transform('count') - 1
        return (t / t_n).values - 0.5

    @staticmethod
    def normalize_series_between_with_minmax(series):
        return 2 * ((series - series.min()) / (series.max() - series.min())) - 1


import os
import pandas as pd

JKP_signals = ['cowc_gr1a', 'oaccruals_at', 'oaccruals_ni', 'taccruals_at', 'taccruals_ni',
               'capex_abn', 'debt_gr3', 'fnl_gr1a', 'ncol_gr1a', 'nfna_gr1a', 'noa_at',
               'aliq_at', 'at_gr1', 'be_gr1a', 'capx_gr1', 'capx_gr2',
               'capx_gr3', 'coa_gr1a', 'col_gr1a', 'emp_gr1', 'inv_gr1', 'inv_gr1a',
               'lnoa_gr1a', 'mispricing_mgmt', 'ncoa_gr1a', 'nncoa_gr1a', 'noa_gr1a',
               'ppeinv_gr1a', 'ret_60_12', 'sale_gr1', 'sale_gr3', 'saleq_gr1',
               'seas_2_5na', 'age', 'aliq_mat', 'at_be', 'bidaskhl_21d', 'cash_at',
               'netdebt_me', 'ni_ivol', 'rd_sale', 'rd5_at', 'tangibility', 'z_score',
               'beta_60m', 'beta_dimson_21d', 'betabab_1260d', 'betadown_252d',
               'chcsho_12m', 'earnings_variability', 'eqnetis_at', 'fcf_me', 'ivol_capm_21d',
               'ivol_capm_252d', 'ivol_ff3_21d', 'ivol_hxz4_21d', 'netis_at', 'ocfq_saleq_std',
               'rmax1_21d', 'rmax5_21d', 'rvol_21d', 'turnover_126d', 'zero_trades_126d',
               'zero_trades_21d', 'zero_trades_252d', 'prc_highprc_252d', 'resff3_12_1',
               'resff3_6_1', 'ret_12_1', 'ret_3_1', 'ret_6_1', 'ret_9_1', 'seas_1_1na',
               'dsale_dinv', 'dsale_drec', 'dsale_dsga', 'niq_at_chg1', 'niq_be_chg1', 'niq_su',
               'ocf_at_chg1', 'qmj_safety', 'ret_12_7', 'sale_emp_gr1', 'saleq_su', 'seas_1_1an',
               'sti_gr1a', 'dolvol_var_126d', 'ebit_bev', 'ebit_sale', 'f_score', 'intrinsic_value',
               'ni_be', 'niq_be', 'o_score', 'ocf_at', 'ope_be', 'ope_bel1',
               'turnover_var_126d', 'at_turnover', 'cop_at',
               'cop_atl1', 'dgp_dsale', 'gp_at', 'gp_atl1', 'mispricing_perf', 'ni_inc8q',
               'niq_at', 'op_at', 'op_atl1', 'opex_at', 'qmj', 'qmj_growth', 'qmj_prof',
               'sale_bev', 'tax_gr1a', 'corr_1260d', 'coskew_21d', 'dbnetis_at',
               'kz_index', 'lti_gr1a', 'ni_ar1', 'pi_nix', 'seas_11_15an', 'seas_11_15na',
               'seas_16_20an', 'seas_16_20na', 'seas_2_5an', 'seas_6_10an', 'seas_6_10na',
               'ami_126d', 'dolvol_126d', 'market_equity', 'prc', 'rd_me',
               'iskew_capm_21d', 'iskew_ff3_21d', 'iskew_hxz4_21d', 'ret_1_0', 'rmax5_rvol_21d',
               'rskew_21d', 'at_me', 'be_me', 'bev_mev', 'debt_me', 'div12m_me', 'ebitda_mev',
               'eq_dur', 'eqnpo_12m', 'eqnpo_me', 'eqpo_me', 'ni_me', 'ocf_me', 'sale_me']

future_return_column_name = 'ret_exc_lead1m'

def pre_process_raw_signals_130(base_folder, save_dir=None, start_date ='19630101', end_date ='20191231',tolerance_for_nan_in_one_row =0.3,max_nb_of_predictor=130):

    # load the data
    JKP_filename = os.path.join(base_folder, 'usa_factor_data.csv')
    # prepare the target directory to save the resulting folder
    if save_dir is None:
        save_dir = base_folder + '/130_signal/'
    os.makedirs(save_dir, exist_ok=True)

    # select the columns of interest
    id_col = ['id', 'eom', 'size_grp']
    cols_read = id_col + JKP_signals + [future_return_column_name]

    # load the csv
    df = pd.read_csv(JKP_filename, usecols=cols_read, parse_dates=['eom'])

    # drop nano
    df = df.loc[df['size_grp'] != 'nano', :]

    # drop low and upper date
    ind = (df['eom'] >= pd.to_datetime(start_date)) & (df['eom'] <= pd.to_datetime(end_date))
    df = df.loc[ind, :].reset_index(drop=True)

    if max_nb_of_predictor > 0:
        # check missing values across the full sample
        # and keep only max_nb_of_predictor with fewer number of missing values
        all = pd.isna(df[JKP_signals]).mean()
        all.name = 'all'
        df = df.loc[~pd.isna(df[future_return_column_name]), :]
        keep_col = list(pd.isna(df[JKP_signals]).mean().sort_values().head(max_nb_of_predictor).index)
        df = df.loc[:, id_col + keep_col + [future_return_column_name]]

    # drop the rows with more than tolerance_for_nan_in_one_row o Nan
    ind = pd.isna(df[keep_col]).mean(1) <= tolerance_for_nan_in_one_row
    df = df.loc[ind, :].reset_index(drop=True)
    df = df.rename(columns={'eom': 'date'})

    # save per size group in pickle to faciltate
    for size in df['size_grp'].unique():
        df.loc[df['size_grp'] == size, :].drop(columns='size_grp').to_pickle(save_dir + size + '.p')

if __name__ == '__main__':
    # example of one of the pre-processing
    pre_process_raw_signals_130(base_folder='work/PRTNR/EPFL/CDM/smalamud/complexmodels/virtualcomplexityeverywhere_data/')



import numpy as np
def smart_eigenvalue_decomposition_covariance(features: np.ndarray):
    """
    Lemma 28: Efficient Eigen value decomposition
    :param n: normalization
    :param features: features used to create covariance matrix times x P
    :param T: Weight used to normalize matrix
    :return: Left eigenvectors PxT and eigenvalues without zeros
    """
    [T, P] = features.shape

    if P > T:
        covariance = features @ features.T /T

    else:
        covariance = features.T @ features / T

    eigval, eigvec = np.linalg.eigh(covariance)
    # eigvec = eigvec[:, eigval > 10 ** (-10)]
    # eigval = eigval[eigval > 10 ** (-10)]

    if P > T:
        # project features on normalized eigenvectors
        eigvec = features.T @ eigvec * (eigval ** (-1 / 2)).reshape(1, -1) / np.sqrt(T)

    return eigval, eigvec



import setuptools

setuptools.setup(
    name='didipack',
    packages=setuptools.find_packages(),  # Automatically discover and include all packages in the package directory
    version='4.0.6',
    license='MIT',
    setup_requires=['wheel'],
    description='Usefull time saver for data-science and finance with python',
    author='Antoine Didisheim',
    author_email='antoinedidisheim@gmail.com',
    url='https://github.com/AntoineDidisheim',
    download_url='https://github.com/AntoineDidisheim/didipack/archive/v0.1.1.tar.gz',
    keywords=['LaTex', 'Finance'],
    install_requires=[
        'numpy',
        'pandas',
        'statsmodels',
        'matplotlib',
        'jinja2',
        'tqdm'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    # If you have package data
    # package_data={
    #     'didipack': ['data/*.csv'],  # example, adjust as needed
    # },
)




import matplotlib.pyplot as plt

class PlotPlus:
    @staticmethod
    def plot_dual_axis(df, col1, col2, label1=None, label2=None, xlabel = 'Date'):
        fig, ax1 = plt.subplots()

        # Use column names as labels if no labels are provided
        if label1 is None:
            label1 = col1
        if label2 is None:
            label2 = col2

        # Plotting col1 on the first y-axis
        color = 'tab:blue'
        ax1.set_xlabel(xlabel=xlabel)
        ax1.set_ylabel(label1, color=color)
        ax1.plot(df.index, df[col1], color=color)
        ax1.tick_params(axis='y', labelcolor=color)

        # Creating a second y-axis to plot col2
        ax2 = ax1.twinx()

        color = 'tab:red'
        ax2.set_ylabel(label2, color=color)
        ax2.plot(df.index, df[col2], color=color)
        ax2.tick_params(axis='y', labelcolor=color)

        fig.tight_layout()
